/* Implementations of CGraph utility functions

   Author:  Timothy Rolfe

   This is predominantly the implementation of the behaviors for
   Deque (the double-ended queue) and MinHeap (the priority queue).
   A small amount of CVertex needs to be implemented as well.

   CVertex::~CVertex(void)
   CVertex::setName(const char*)

   Deque::~Deque(void);
   Deque::push(const CPair&);
   Deque::enqueue(const CPair&);
   Deque::remove (CPair&);

   MinHeap::~MinHeap(void);
   MinHeap::enter (const CPair&, const CEdge &);
   MinHeap::remove(CPair&, CEdge &);
   MinHeap::upHeap (int P);
   MinHeap::downHeap(int P);
*/

#include "GrUtil.h"
#include <string.h>     // strlen and strcpy in CVertex::setName
#include <assert.h>     // assert in CVertex::setName
#include <iostream.h>   // cerr message on remove from empty structure

/***************************** CVertex *******************************/

CVertex::~CVertex(void)
{  delete [] m_Name;  }

void CVertex::setName(const char *N)
{
   if ( m_Name )
      delete [] m_Name;
   m_Name = new char [ 1 + strlen(N) ];
   assert (m_Name);
   strcpy (m_Name, N);
}

/***************** Deque of ordered pairs *****************/

Deque::~Deque(void)
{
   while ( m_Front )
   {
      DQnode *Hold = m_Front;
      m_Front = m_Front->m_Next;
      delete Hold;
   }
}

/**********************************************************/

void Deque::push(const CPair &V)     // Stack-discipline entry
{
   m_Front = new DQnode(V, m_Front);
   if ( m_Back == NULL )
      m_Back = m_Front;                // Update for queue use
}

/**********************************************************/

void Deque::enqueue(const CPair &V)  // Queue-discipline entry
{
   if ( m_Front == NULL )
      m_Front = m_Back = new DQnode(V, NULL);
   else
      m_Back = m_Back->m_Next = new DQnode(V, NULL);
}

/**********************************************************/

bool Deque::remove (CPair &V)         // Common removal point
{
   DQnode *Hold = m_Front;

   if ( ! Hold )
      return false;

   V = m_Front->m_Edge;
   m_Front = m_Front->m_Next ;
   if ( ! m_Front )                    // Clean up for queue use
      m_Back = NULL;
   delete Hold;

   return true;
}

/***************** MinHeap of ordered pairs *****************/

// NOTE: the heap is implemented to use [1] as its root.

MinHeap::MinHeap(int Size)           // Create the array
:  m_N(0), m_Size(Size)
{
// Use the alloc family to allow use of realloc when resizing
   m_Heap = (MinHnode*) calloc (Size+1, sizeof *m_Heap);
   m_Heap[0].m_Dist = -1;              // Sentinel value
}

/**********************************************************/

MinHeap::~MinHeap(void)
{
   free (m_Heap);  // Allocated by the alloc family
}

/**********************************************************/

void MinHeap::enter (const CPair &Edge, int Dist)
{
   if ( m_N == m_Size ) // About to overflow:  expand the array
   {
      m_Size += 5;      // or whatever increment feels ok.
      m_Heap = (MinHnode*) realloc ( m_Heap, (m_Size+1)*sizeof(*m_Heap) );
      if ( m_Heap == NULL )
      {  cerr << "Memory overlow.  Abort!" << endl;  exit(-1);  }
   }
// Since [1] is the root, increment m_N, then use as subscript
   m_N++;
   m_Heap[m_N].m_Edge = Edge;
   m_Heap[m_N].m_Dist = Dist;
   upHeap(m_N);
}

/**********************************************************/

bool MinHeap::remove(CPair &Edge, int &Dist)
{
   if ( m_N < 1 )
      return false;

   Edge = m_Heap[1].m_Edge;
   Dist = m_Heap[1].m_Dist;
   m_Heap[1] = m_Heap[m_N--];
   downHeap(1);

   return true;
}

/**********************************************************/

void MinHeap::upHeap  (int C)    // Correct the heap from C upwards
{
   MinHnode Hold = m_Heap[C];
   int P = C/2;

// m_Heap[0] has a sentinel value forcing loop exit when P == 0.
   while ( Hold.m_Dist < m_Heap[P].m_Dist )
   {
      m_Heap[C] = m_Heap[P];
      C = P;
      P = C/2;
   }
   m_Heap[C] = Hold;
}

/**********************************************************/

void MinHeap::downHeap(int P)    // Correct the heap from P downwards
{
   MinHnode Hold = m_Heap[P];
   int C = 2*P;

   while (C <= m_N)
   {
      if ( C < m_N && m_Heap[C].m_Dist > m_Heap[C+1].m_Dist )
         C++;
      if ( ! (m_Heap[C].m_Dist < Hold.m_Dist) )
         break;
      m_Heap[P] = m_Heap[C];
      P = C;
      C = 2*P;
   }
   m_Heap[P] = Hold;
}

/**********************************************************/
