/* Utility Classes for CGraph

   Author:  Timothy Rolfe

   CVertex  --- the cities
   CEdge    --- the roads
   CPair    --- ordered pairs (to represent an edge)
   Deque    --- double-ended queue (stack AND queue in one package)
   MinHeap  --- priority queue (minimum entry found at root)
*/

#ifndef GRUTILITY_H
#define GRUTILITY_H

#include <stdlib.h>     // Somebody has to define NULL

#ifdef __TURBOC__       // I.e., my Turbo C++ version 3.1 compiler
enum bool {  false,  true  };
#endif

/***************** Vertex --- a city in the map *****************/

class CVertex // Implementation file will do setName and ~CVertex
{
   public:
      CVertex(char *N = "") : m_Name(NULL), m_Flag(0)
      {  setName(N);  }

     ~CVertex(void);

      void setName(const char *N);
      const char* getName(void) const { return m_Name;  }

      void setFlag(int N)       { m_Flag = N;  }
      int  getFlag(void)  const { return m_Flag;  }
      void incFlag(void)        { ++m_Flag;  }
      void decFlag(void)        { --m_Flag;  }

      void setDistance(int D)       { m_Distance = D; }
      int  getDistance(void)  const { return m_Distance; }

   private:
      char *m_Name;
      int   m_Flag;     // Pending, Unvisited, or Visited in traversal
      int   m_Distance; // Distance from start in Dijkstra's algorithm
};

/***************** Edge --- a road in the map *****************/

class CEdge   // ALL behaviors are defined here
{  public:

      CEdge(int Miles=0) : m_Distance(Miles) { }

//    Accesses are just interested in the distance
      int length(void) const {  return m_Distance;  }

//    AllPaths uses " *= -1 " on an edge
      CEdge operator *= ( int Mult )
      {  return (m_Distance *= Mult);  }

   private:
      int m_Distance;
};

/**************** Two-ple (ordered pair of number) ****************/

// NOTE:  behaviors of default assignment operator and
//        default copy constructor are acceptable

class CPair  // ALL behaviors are defined here
{
   public:

      CPair(int V0 = 0, int V1 = 0)
      {  set(V0, V1);  }

      void set(int V0, int V1)
      {  m_First=V0;  m_Second=V1;  }

      int  get1(void) const {  return m_First;  }
      int  get2(void) const {  return m_Second; }

   private:
      int m_First, m_Second;
};

/***************** Deque of 2-ples (ordered pairs) *****************/

class Deque // Double-ended queue (stack AND queue behaviors)
{
   public:
      Deque(void) : m_Front(NULL), m_Back(NULL) { }
     ~Deque(void);

      void push(const CPair&);     // Stack discipline
      void enqueue(const CPair&);  // Queue discipline
      bool remove (CPair&);        // Single removal point
      int  empty(void) { return m_Front == NULL; }

   private:

      struct DQnode
      {
         CPair   m_Edge;
         DQnode *m_Next;

         DQnode(const CPair &V, DQnode *Nxt = NULL)
         :  m_Edge(V), m_Next(Nxt)  {  }
      };

      DQnode *m_Front, *m_Back;
};

/***************** MinHeap of 2-ples (ordered pairs) *****************/

class MinHeap // Priority queue with smallest entry at root
{
   public:

      MinHeap(int Size = 0);
     ~MinHeap(void);

      void enter (const CPair&, int);
      bool remove(CPair&, int &);
      int  empty(void) { return m_N == 0; }

   private:

      struct MinHnode
      {
         CPair m_Edge;
         int   m_Dist;
      };

      void  upHeap (int P);  // Correct the heap from P upwards
      void downHeap(int P);  // Correct the heap from P downwards

      MinHnode *m_Heap;      // Dynamic array
      int m_Size;            // Physical size (number of allocated cells)
      int m_N;               // Logical size (number of entries)
};

#endif
