/* Graph.cpp

   Author:  Timothy Rolfe

   Implementation of class CGraph member functions.
*/
#include "Graph.h"
#include <string.h>     // strlen etc.
#include <stdlib.h>     // atoi
#include <assert.h>
#include <ctype.h>      // isspace

/**********************************************************/

CGraph::~CGraph(void)
{
   for ( int r = 0; r < m_Size; r++ )
   {
      for ( int c = 0; c < m_Size; c++ )
         if (m_Road[r][c])
         {
            delete m_Road[r][c];
            m_Road[r][c] = m_Road[c][r] = NULL;
         }
      free (m_Road[r]); // Obtained by calloc, not new
   }

   delete [] m_Road;    // Obtained by new, not calloc

   delete [] m_City;
// And ~CVertex will give back ALL the city name strings
}// end destructor

/**********************************************************/

// Case-insensitive comparison (some compilers do not have stricmp)
int  CGraph::stricmp(const char *L, const char *R)
{
   while ( *L && *R && *L == *R )
   {  L++;  R++; }
   return *L - *R;
}

/**********************************************************/

// Return subscript for Name in the m_City array --- or -1 for failure
int  CGraph::findCity(char *Name)
{
   for ( int k = 0; k < m_Size; k++ )
      if ( CGraph::stricmp(Name, m_City[k].getName()) == 0 )
         return k; // Success!
   return -1;      // Failure flag
}// end findCity()

/**********************************************************/

// Major function:  initialize the graph from the input stream
void CGraph::fillGraph(istream &Inp)
{
   char LineIn[255];
   int  miles;
   int  row,
        col;

// Insure that fillGraph is called only ONCE:
   assert ( m_Road == NULL );
   
// First record:  number of cities (as a full line)
   Inp.getline(LineIn, 255);
   m_Size = atoi(LineIn);
   assert (m_Size > 0);           // Some entries

   m_City = new CVertex [m_Size];
   assert (m_City != NULL);

   m_Road = new CEdge** [m_Size];
   assert (m_Road != NULL);

// First part:  list of city names
   for ( row = 0; row < m_Size; row++ )
   {
//    Allocate the row AND fill it with NULL pointers
      m_Road[row] = (CEdge**) calloc (m_Size, sizeof(CEdge**));
      Inp.getline(LineIn, 255);
      m_City[row].setName(LineIn);
   }// end for row...

// Remaining portion:  double-quoted names followed by distance
// Enter into BOTH table locations
   Inp >> *LineIn;      // First character
   while ( !Inp.eof() )
   {//Some Code Warrior versions have a problem on eof; hence:
      if ( Inp.fail() )
      {  cerr << "Input stream has failed after " << m_NRoads << " roads."
              << endl;
         break;
      }
//    Reading in potentially double-quoted text
      if ( *LineIn == '"' )
         Inp.getline(LineIn, 255, '"');
      else if (isspace(Inp.peek()))
         LineIn[1] = '\0';
      else
         Inp >> LineIn+1;

      row = findCity(LineIn);
      if ( row < 0 )
      {
         cout << "Unknown 1st city " << LineIn
              << ", discarding rest of record: ";
         Inp.getline(LineIn, 255);
         cout << LineIn << endl;
         continue;
      }// end if row<0
      
      Inp >> *LineIn;
      if ( *LineIn == '"' )
         Inp.getline(LineIn, 255, '"');
      else if (isspace(Inp.peek()))
         LineIn[1] = '\0';
      else
         Inp >> LineIn+1;

      col = findCity(LineIn);
      if ( col < 0 )
      {
         cout << "Road from " << m_City[row].getName()
              << ", destination " << LineIn
              << " unknown, discarding rest of record: ";
         Inp.getline(LineIn, 255);
         cout << LineIn << endl;
         continue;
      }// end if col<0

      Inp.getline(LineIn, 255);
      miles = atoi(LineIn);
      m_Road[row][col] = m_Road[col][row]
                       = new CEdge(miles);

      m_NRoads++;

      Inp >> *LineIn;   // First character from the next record
   }// end while !eof

   m_Sequence = new CPair [m_NRoads];   // Sequence of edges
   assert (m_Sequence != NULL);
}//end fillGraph()

/**********************************************************/

int CGraph::distance(char *begin, char *finish)
{
   int rtnValue = 0;                   // Flag value for invalid

   int r = findCity(begin),
       c = findCity(finish);

   if ( m_Road[r][c] )
      rtnValue = m_Road[r][c]->length();

   return rtnValue;
}//end distance()

/**********************************************************/

void  CGraph::clearFlags(void)    // Clear traversal flags
{
   static const int MAX_INT = ~( 1 << (8*sizeof(int)-1) );

   for (int k = 0; k < m_Size; k++)
   {
      m_City[k].setFlag(Unvisited);
      m_City[k].setDistance(MAX_INT);
   }
   m_NextEdge = 0;
}

/**********************************************************/

void CGraph::edgeList   ( ostream &Fout )
{
   int  Sum = 0, Leg;

   for ( int k = 0; k < m_NextEdge; k++ )
   {  int Src = m_Sequence[k].get1(),
          Dst = m_Sequence[k].get2();

      if ( Src < 0 )
         Fout << "Starting from " << m_City[Dst].getName() << endl;
      else
      {
         Leg = m_Road[Src][Dst]->length();
         Sum += Leg;
         Fout << m_City[Src].getName() << " to "
              << m_City[Dst].getName() << " ("
              << Leg << ")" << endl;
      }
   }
   if ( &Fout == &cout )
   {
      cout << "\nTotal distance: " << Sum 
           << "   Press ENTER to continue: ";
      cin.ignore(255, '\n');
      cout  << endl;
   }
   else
      Fout << "\nTotal distance: " << Sum << endl;
}

/*********************************************/

// Depth-First Traversal --- Recursive

void CGraph::depthRecur ( char *Start )
// Entry point
{//Decode start subscript
   int  First = findCity(Start);

   if (First < 0)
   {  cerr << "Unknown starting point " << Start
           << " --- abort!" << endl;
      return;
   }

   clearFlags();
   depthRecur ( -1, First );

// Take care of possible unconnected components
   for ( First = 0; First < m_Size; First++ )
      if (m_City[First].getFlag() == Unvisited)
         depthRecur ( -1, First );
}

void CGraph::depthRecur ( int Src, int Dst )
// Implementation code
{  CPair Edge(Src, Dst);

   m_Sequence[m_NextEdge++] = Edge;
// Save position in the list
   m_City[Dst].setFlag(m_NextEdge);

   Src = Dst;
   for ( Dst = 0; Dst < m_Size; Dst++ )
      if ( m_Road[Src][Dst] )
      {  int Flag = m_City[Dst].getFlag();

         if (Flag == Unvisited )
            depthRecur (Src, Dst);
      }
}

/*********************************************/

// Depth-First Traversal --- Iterative

void CGraph::depth1st ( char* Start )
// Entry point
{//Decode start subscript
   int First = findCity(Start);

   if (First < 0)
   {  cerr << "Unknown starting point " << Start
           << " --- abort!" << endl;
      return;
   }

   clearFlags();
   depth1st ( First );

// Take care of possible unconnected components
   for ( First = 0; First < m_Size; First++ )
      if (m_City[First].getFlag()==Unvisited)
         depth1st ( First );
}

void CGraph::depth1st ( int First )
// Implementation code
{  CPair Edge;
   Deque Work;

   Edge.set(-1, First);      // Pseudo-edge for entry point
   Work.push(Edge);

// remove returns "true" if the operation succeeds
   while ( Work.remove(Edge) )
   {  int Src, Dst;

      Src = Edge.get1();
      Dst = Edge.get2();
      if ( m_City[Dst].getFlag() != Unvisited )
         continue;

      m_Sequence[m_NextEdge++] = Edge;
//    Save position in the list
      m_City[Dst].setFlag(m_NextEdge);

      Src = Dst;
      for ( Dst = m_Size; --Dst >= 0; )
         if ( m_Road[Src][Dst] )
         {  int Flag = m_City[Dst].getFlag();
            Edge.set(Src, Dst);

            if (Flag == Unvisited )
            {
               Work.push(Edge);
            }
         }
   }
}

/**********************************************************/

// Breadth-First Traversal

void CGraph::breadth1st ( char* Start )
// Entry point
{//Decode start subscript
   int First = findCity(Start);

   if (First < 0)
   {  cerr << "Unknown starting point " << Start
           << " --- abort!" << endl;
      return;
   }

   clearFlags();
   breadth1st ( First );

// Take care of possible unconnected components
   for ( First = 0; First < m_Size; First++ )
      if (m_City[First].getFlag()==Unvisited)
         breadth1st ( First );
}

void CGraph::breadth1st ( int First )
// Implementation code
{  CPair Edge;
   Deque Work;

   Edge.set(-1, First);      // Pseudo-edge for entry point
   Work.enqueue(Edge);

// remove returns "true" if the operation succeeds
   while ( Work.remove(Edge) )
   {  int Src, Dst;

      Src = Edge.get1();
      Dst = Edge.get2();
      if ( m_City[Dst].getFlag() != Unvisited )
         continue;

      m_Sequence[m_NextEdge++] = Edge;
//    Save position in the list
      m_City[Dst].setFlag(m_NextEdge);

      Src = Dst;
      for ( Dst = 0; Dst < m_Size; Dst++ )
         if ( m_Road[Src][Dst] )
         {  int Flag = m_City[Dst].getFlag();
            Edge.set(Src, Dst);

            if ( Flag == Unvisited )
            {
               Work.enqueue(Edge);
            }
         }
   }
}

/**********************************************************/

void CGraph::Prim ( char* Start )                // Entry point
{//Decode start subscript
   int First = findCity(Start);

   if (First < 0)
   {
      cerr << "Unknown starting point " << Start
           << " --- abort!" << endl;
      return;
   }

   clearFlags();

   Prim ( First );

// Take care of possible unconnected components
   for ( First = 0; First < m_Size; First++ )
      if ( m_City[First].getFlag() == Unvisited )
         Prim ( First );
}

// Minimum Spanning Tree Generation (Prim's Algorithm)

void CGraph::Prim ( int First )
// Implementation code
{  CPair    Edge;
   MinHeap  Work;
   int      Distance = 0;

   Edge.set(-1, First);      // Pseudo-edge for entry point
   Work.enter(Edge, Distance);

// remove returns "true" if the operation succeeds
   while ( Work.remove(Edge, Distance) )
   {  int Src, Dst;

      Src = Edge.get1();
      Dst = Edge.get2();
      if ( m_City[Dst].getFlag() != Unvisited )
         continue;
   
      m_Sequence[m_NextEdge++] = Edge;
//    Save position in the list
      m_City[Dst].setFlag(m_NextEdge);

      Src = Dst;
      for ( Dst = 0; Dst < m_Size; Dst++ )
         if ( m_Road[Src][Dst] )
         {  int Flag = m_City[Dst].getFlag();
            Edge.set(Src, Dst);
            Distance = m_Road[Src][Dst]->length();

            if ( Flag == Unvisited )
            {
               Work.enter(Edge, Distance);
            }
         }
   }
}

/**********************************************************/

void CGraph::Dijkstra ( char* Start )            // Entry point
{//Decode start subscript
   int First = findCity(Start);

   if (First < 0)
   {
      cerr << "Unknown starting point " << Start
           << " --- abort!" << endl;
      return;
   }

   clearFlags();

   Dijkstra ( First );

// Take care of possible unconnected components
   for ( First = 0; First < m_Size; First++ )
      if ( m_City[First].getFlag() == Unvisited )
         Dijkstra ( First );
}

void CGraph::Dijkstra ( int First )
// Implementation code
{  CPair    Edge;
   MinHeap  Work;
   int      Distance = 0;

// Zero distance to the starting city, of course!
   m_City[First].setDistance(0);

   Edge.set(-1, First);      // Pseudo-edge for entry point
   Work.enter(Edge, Distance);

// remove returns "true" if the operation succeeds
   while ( Work.remove(Edge, Distance) )
   {  int Src, Dst;

      Src = Edge.get1();
      Dst = Edge.get2();
      if ( m_City[Dst].getFlag() != Unvisited )
         continue;
   
      m_Sequence[m_NextEdge++] = Edge;
//    Save position in the list
      m_City[Dst].setFlag(m_NextEdge);

      Src = Dst;
      for ( Dst = 0; Dst < m_Size; Dst++ )
         if ( m_Road[Src][Dst] )
         {  int Flag = m_City[Dst].getFlag();
            Edge.set(Src, Dst);
            Distance = m_Road[Src][Dst]->length();
            Distance += m_City[Src].getDistance();
//          setFlags initializes distances to MAX_INT
            if ( Flag == Unvisited &&
                 m_City[Dst].getDistance() > Distance)
            {//DO enter the edge and update the shorter distance
               m_City[Dst].setDistance(Distance);
               Work.enter(Edge, Distance);
            }
         }
   }
}

/**********************************************************/
