/* Graph.h

   Author:  Timothy Rolfe
            loosely based on Robert Sedgewick's Algorithms in C++

   Graph classes in use:
   (1) CGraph   --- the map itself
   (2) CVertex  --- the cities in the map (in GrUtility.h)
   (3) CEdge    --- the roads connecting them (in GrUtility.h)

   The graph is represented by a vector of objects representing the
   vertices (cities in the map), and a matrix of objects representing
   the edges (roads connecting the cities).  These are considered to
   be parallal arrays:  subscript k refers to a particular city in
   the vertex array, and refers to the same city as row and column
   subscripts in the edge matrix.

   Each vertex contains a string with the town name and a int flag field
   for traversal.  Some algorithms may require additional entries
   (for instance, those required for the single-source shortest path
   determination by Dijkstra's algorithm).

   The two-dimensional array contains pointers to objects representing
   the edges (roads in the map).  A NULL pointer indicates that there
   is no edge connecting the two vertices.  Each edge object contains
   (presently) just an integer giving the weight (length of the road
   between the two cities).  Conceivably it might have a string giving
   the highway name.

   Presumed input:
   (1) Number of vertices (cities)
   (2) THAT NUMBER of lines giving the city names
   (3) Remaining entries:  three fields:
       two vertex designations (city names) and the weight (road length).
   EOF marks the end of edge information

   Utility classes in use (in GrUtility.h):
   (1) CPair   --- an ordered pair to represent an edge (source,dest)
   (2) DeQueue --- double-ended queue, to be used either in stack
                   mode (depth-first) or queue mode (breadth-first)
   (3) MinHeap --- used by Prim's and Dijkstra's algorithms
   ___________________________________________________________

   Details about class CGraph

   Supported public functions:

      constructor --- default destructor only (empty graph)
      destructor --- LOTS of dynamic memory is used!
      fillGraph receives the istream with the data
      nCities() and nRoads() --- "get" functions
      validCity(char*) --- test whether the city is in the graph
      distance(char*, char*) --- distance between two cities if there
                                 is a road, else zero as error flag

   Assorted traversal functions and their utilities:

      depthRecur  Recursive implementation
      depth1st    User-stack-driven implementation
      breadth1st  Nearly identical, but with a queue, not a stack
      Prim        Prim's algorithm for minimum spanning tree
      Dijkstra    Dijkstra's algorithm for 1-source shortest path

      edgeList    List the edges used in the spanning tree

   Attributes

      Weight matrix: 2-D array of pointers to CEdge objects
      Dynamic array of vertices (holding city names)
      Dynamic array holding the sequence of edges
      Number of cities in the graph
      Number of roads
*/
#ifndef GRAPH_H
#define GRAPH_H

#include <iostream.h>
#include "GrUtil.h"     // Assorted utility classes for CGraph

/********************* CGraph Class Declaration *********************/

class CGraph
{
   public:
      CGraph(void) : m_Size(0), m_NRoads(0), m_Road(NULL)
      { /* Nothing left to do */ }

      ~CGraph(void);

      void fillGraph(istream&);        // Read in ALL the graph data

      int  nCities(void) { return m_Size; }
      int  nRoads(void)  { return m_NRoads; }

      bool validCity(char *V)
      {  return findCity(V) >= 0;  }

      int distance(char *, char *);    // Edge length

//    Traversal routines --- parameter is the start city

//    Each of the following uses a private member function
      void depthRecur ( char* Start ); // Recursive depth-first
      void depth1st   ( char* Start ); // Iterative depth-first
      void breadth1st ( char* Start ); // Breadth-first
      void Prim       ( char* Start ); // Minimum spanning tree
      void Dijkstra   ( char* Start ); // Single-source shortest paths

//    List the edges in the resulting spanning tree
      void edgeList   ( ostream & );

   protected:
      enum { Unvisited=0 };  // Only used in CGraph

//    Some compilers do not provide case-insensitive comparisons
      int  stricmp (const char*, const char*);
      int  findCity(char*);  // Subscript of city
      void clearFlags(void); // Clear traversal flags

//    Called by the respective public member functions
      void depthRecur ( int Source, int Dest );
      void depth1st   ( int Source );
      void breadth1st ( int Source );
      void Prim       ( int Source );
      void Dijkstra   ( int Source );

      CEdge**  *m_Road;      // Dynamic 2-d array of pointers
      CVertex  *m_City;      // Dynamic array of vertices
      CPair    *m_Sequence;  // Sequence of edges
      int       m_Size;      // Number of cities
      int       m_NRoads;
      int       m_NextEdge;  // Used in filling m_Sequence
};

#endif
