/* Exponential algorithms:  examine ALL paths

   The starting point is allPaths --- which is based
   on Program 8-4 in Antonakos and Mansfield, "Practical
   Data Structures Using C/C++", pp. 268-69.

   allPaths is based on marking the used EDGES, prohibiting
   traversal of any _edge_ more than once.

   Revisions and extensions to Antonakos and Mansfield by
   Timothy Rolfe
   
   simplePaths --- prohibit visiting any city more than once
   Hamiltonian --- a tour (simple path of all cities ending with
                   an edge back to the initial city)
   travelSales --- based on preprocessor setting, this can relax
                   the restriction on revisiting vertices in the
                   circuit.  It always, however, reports ONLY
                   a sequence with minimum total distance (or, by
                   a preprocessor line change, maximum distance).
   minSeq      --- Generate, update, or report the sequence found
                   by travelSales
*/
//Preprocessor settings to tune behaviors in travelSales/minSeq
//#define HAMILTONIAN     // I.e., in travelSales()
//#define MAX_CIRCUIT     // ditto
//#define TERSE           // Disable reports in minSeq()

#include "GraphAll.h"

/**********************************************************/

// Entry point:  character strings giving start and end
void CGraphAll::allPaths(char *Start, char *End)
{
   int Src = findCity(Start),
       Dst = findCity(End);

// Zero out the edge list
   m_NextEdge = 0;
   if ( Src >= 0 && Dst >= 0 )
      allPaths(Src, Dst);
   else
      cerr << "Unknown city.  No paths." << endl;
}

// Display ALL paths between Src and Dst (i.e., EDGES marked as used)
// Based on Program 8-4 in J. L. Antonakos and K. C. Mansfield,
// "Practical Data Structures Using C/C++" (1999), pp. 268-69.
void CGraphAll::allPaths(int Src, int Dst)
{
   int k;
   int ChkWt = ( m_Road[Src][Dst] ? m_Road[Src][Dst]->length() : 0);

   if ( ChkWt > 0 )               // Does the path terminate at Dst?
   {
//    Enter final edge into the sequence
      m_Sequence[m_NextEdge++] = CPair(Src, Dst);
      edgeList(cout);
      m_NextEdge--;     // Remove that edge from the sequence
   }
   for ( k = 0; k < m_Size; k++ ) // Check all edges out of Src
   {
      ChkWt = ( m_Road[Src][k] ? m_Road[Src][k]->length() : 0);
      if ( ChkWt > 0 )            // I.e., This is an unused edge
      {
//       Enter this edge into the sequence
         m_Sequence[m_NextEdge++] = CPair(Src, k);
//       Note:  [Src][k] AND [k][Src] point at the SAME object
         *m_Road[Src][k] *= -1;   // Mark edge as used
         allPaths ( k, Dst );     // Recursive call
         *m_Road[Src][k] *= -1;   // Mark edge as UNused
         m_NextEdge--;            // Remove this edge
      }
   }
}

/**********************************************************/

// Entry point:  character strings giving start and end
void CGraphAll::simplePaths(char *Start, char *End)
{
   int Src = findCity(Start),
       Dst = findCity(End);

// We will be using vertex marking as well as edge marking
   clearFlags();
   m_NextEdge = 0;
   if ( Src >= 0 && Dst >= 0 )
   {
//    Mark the starting point as visited.
      m_City[Src].setFlag(1);
      simplePaths(Src, Dst);
   }
   else
      cerr << "Unknown city.  No paths." << endl;
}

// Display all SIMPLE paths between Src and Dst (add vertex marking)
void CGraphAll::simplePaths(int Src, int Dst)
{
   int k;
   int ChkWt = ( m_Road[Src][Dst] ? m_Road[Src][Dst]->length() : 0);

   if ( ChkWt > 0 )               // Does the path terminate at Dst?
   {
//    Mark termination as visited
      m_City[Dst].setFlag(1);
      m_Sequence[m_NextEdge++] = CPair(Src, Dst);
      edgeList(cout);
      m_NextEdge--;
   }
   for ( k = 0; k < m_Size; k++ ) // Check all edges out of Src
   {
      ChkWt = ( m_Road[Src][k] ? m_Road[Src][k]->length() : 0);
//    This is an unused edge, and the destination is unvisited
      if ( ChkWt > 0 && m_City[k].getFlag() == Unvisited )
      {
         m_City[k].setFlag(1);
         m_Sequence[m_NextEdge++] = CPair(Src, k);
//       Note:  [Src][k] AND [k][Src] point at the SAME object
         *m_Road[Src][k] *= -1;   // Mark edges as used
         simplePaths ( k, Dst );  // Recursive call
         *m_Road[Src][k] *= -1;   // Mark edges as UNused
         m_NextEdge--;            // Remove from path
         m_City[k].setFlag(Unvisited);
      }
   }
}

/**********************************************************/

// Entry point
void CGraphAll::Hamiltonian(char *Start)
{
   int Src = findCity(Start);

   m_NextEdge = 0;
   if ( Src >= 0 )
   {
      m_City[Src].setFlag(1);
      clearFlags();
      Hamiltonian(Src, Src);
   }
   else
      cerr << "Unknown city.  No paths." << endl;
}

// Simple path among all the cities that ends up back at the source
void CGraphAll::Hamiltonian(int Src, int Dst)
{
   int k;
   int ChkWt = ( m_Road[Src][Dst] ? m_Road[Src][Dst]->length() : 0);

   if ( ChkWt > 0 )               // Does the path terminate at Dst?
   {
      m_City[Dst].setFlag(1);
//    ALL cities included if one fewer edge than there are cities
//    Note that the final edge is added within the if block.
      if ( m_NextEdge == m_Size-1 )
      {
         m_Sequence[m_NextEdge++] = CPair(Src, Dst);
         edgeList(cout);
         m_NextEdge--;
      }
   }
   for ( k = 0; k < m_Size; k++ ) // Check all edges out of Src
   {
      ChkWt = ( m_Road[Src][k] ? m_Road[Src][k]->length() : 0);
//    This is an unused edge, and the destination is unvisited
      if ( ChkWt > 0 && m_City[k].getFlag() == Unvisited )
      {
         m_City[k].setFlag(1);
         m_Sequence[m_NextEdge++] = CPair(Src, k);
//       Note:  [Src][k] AND [k][Src] point at the SAME object
         *m_Road[Src][k] *= -1;   // Mark edges as used
         Hamiltonian ( k, Dst );  // Recursive call
         *m_Road[Src][k] *= -1;   // Mark edges as UNused
         m_NextEdge--;            // Remove from path
         m_City[k].setFlag(Unvisited);
      }
   }
}

/**********************************************************/
 
void CGraphAll::travelSales( char* Start )
{  int Src = findCity(Start);

   m_NextEdge = 0;
   if ( Src >= 0 )
   {  m_City[Src].setFlag(1);
      clearFlags();
      travelSales(Src, Src);
      cout << "\nMinimum sequence found among " << m_PathsChecked
           << " checked:\n\n";
//    Output the minimum sequence actually found
      minSeq(-1, cout);
   }
   else
      cerr << "Unknown city.  No paths." << endl;
} // end public CGraphAll::travelSales

void CGraphAll::travelSales(int Src, int Dst)
{  int k;
   int ChkWt = ( m_Road[Src][Dst] ? m_Road[Src][Dst]->length() : 0);

   if ( ChkWt > 0 )               // Does the path terminate at Dst?
   {  bool Found = true;

      m_City[Dst].incFlag();
      for ( k = 0; k < m_Size; k++ )
         Found = Found && m_City[k].getFlag();
      if ( Found )
      {  m_Sequence[m_NextEdge++] = CPair(Src, Dst);
         minSeq(0);
         m_NextEdge--;
      } // end if (Found)
   } // end if ( ChkWt ...
   for ( k = 0; k < m_Size; k++ ) // Check all edges out of Src
   {  ChkWt = ( m_Road[Src][k] ? m_Road[Src][k]->length() : 0);

#ifdef HAMILTONIAN      /* That is, do NOT revisit vertices */
//    This is an unused edge, and the destination is unvisited
      if ( ChkWt > 0 && m_City[k].getFlag() == Unvisited )
#else
      if ( ChkWt > 0 )            // This is an unused edge.
#endif
      {  m_City[k].incFlag();     // Count UP on flag use
         m_Sequence[m_NextEdge++] = CPair(Src, k);
//       Note:  [Src][k] AND [k][Src] point at the SAME object
         *m_Road[Src][k] *= -1;   // Mark edges as used
         travelSales ( k, Dst );  // Recursive call
         *m_Road[Src][k] *= -1;   // Mark edges as UNused
         m_NextEdge--;            // Remove from path
         m_City[k].decFlag();     // Back off on flag use
      } // end if (ChkWt ...
   } // end for (k=0 ...
} // end private CGraphAll::travelSales

// Avoid colliding with possible library function abs
inline int Abs (int V)
{  return ( V < 0 ? -V : V );  }

//    Options:  -1 to report, +1 to initialize, 0 to update
void CGraphAll::minSeq ( int Opt, ostream &O )
{  static CPair *Working = NULL;
   static int   Size;
   static int   MinWeight = RAND_MAX;
   int  CurrentWeight = 0;
   int  k;

   switch (Opt)
   {//Report:  copy info into m_Sequence and m_NextEdge; use edgeList
   case -1:  for ( k = 0; k < Size; k++ )
              { m_Sequence[k] = Working[k];  }
             m_NextEdge = Size;
             edgeList(O);
             break;

   case 0:   // Initialize/Update saved info based on m_Sequence etc.
   case +1:  bool Init = (Working == NULL || Opt == +1 );

             if ( Init )
             {  if ( Working != NULL )
                   delete [] Working;
                Working = new CPair [ m_NRoads];
                m_PathsChecked = 0;    // private data member
             } // end if (Init)
             m_PathsChecked++;
//           Compute the length of the current path
             for ( k = 0; k < m_NextEdge; k++ )
             {  int Src = m_Sequence[k].get1(),
                    Dst = m_Sequence[k].get2();

                if ( Src >= 0 )   // Src == -1 for initial pseudo-edge
                   CurrentWeight += Abs(m_Road[Src][Dst]->length());
             } // end for ( k = 0 ...
//           Update saved information on Init or a new minimum
//           Accept MOST RECENT sequence with this MinWeight
#ifdef MAX_CIRCUIT
             if ( Init || CurrentWeight >= MinWeight )
#else  // we must be doing the MIN circuit.
             if ( Init || CurrentWeight <= MinWeight )
#endif
             {  Size = m_NextEdge;
                for ( k = 0; k < Size; k++ )
                   Working[k] = m_Sequence[k];
#ifndef TERSE // Verbose mode reports each extreme weight change
                if ( MinWeight != CurrentWeight )
                   cout << "Current minimum weight:  "
                        << CurrentWeight << endl;
#endif
                MinWeight = CurrentWeight;
             } // end if ( Init ...
   } // end switch(Opt)
} // end CGraphAll::minSeq


/**********************************************************/

// Override the default --- absolute values of the edge weights,
// since negative weights are flagging in-use edges (that is,
// all but the last one we would report).
void CGraphAll::edgeList   ( ostream &Fout )
{
   int  Sum = 0, Leg;

   for ( int k = 0; k < m_NextEdge; k++ )
   {  int Src = m_Sequence[k].get1(),
          Dst = m_Sequence[k].get2();

      if ( Src < 0 )
         Fout << "Starting from " << m_City[Dst].getName() << endl;
      else
      {
         Leg = Abs(m_Road[Src][Dst]->length());
         Sum += Leg;
         Fout << m_City[Src].getName() << " to "
              << m_City[Dst].getName() << " ("
              << Leg << ")" << endl;
      }
   }
   if ( &Fout == &cout )
   {
      cout << "\nTotal distance: " << Sum 
           << "   Press ENTER to continue: ";
      cin.ignore(255, '\n');
      cout  << endl;
   }
   else
      Fout << "\nTotal distance: " << Sum << endl;
}

/**********************************************************/
